@@ -15,13 +15,15 @@
 import collections
 import netaddr
 
+from neutron_lib.api.definitions import portbindings
+from neutron_lib.api.definitions import provider_net as pnet
 from neutron_lib.api import validators
 from neutron_lib import constants as const
 from neutron_lib import exceptions as n_exc
+from neutron_lib.plugins import directory
 from oslo_config import cfg
 from oslo_db import exception as os_db_exc
 from oslo_log import log
-import six
 
 from neutron.callbacks import events
 from neutron.callbacks import registry
@@ -29,10 +31,7 @@ from neutron.callbacks import resources
 from neutron.common import utils as n_utils
 from neutron import context as n_context
 from neutron.db import provisioning_blocks
-from neutron.extensions import portbindings
 from neutron.extensions import portsecurity as psec
-from neutron.extensions import providernet as pnet
-from neutron import manager
 from neutron.plugins.common import constants as plugin_const
 from neutron.plugins.ml2 import driver_api
 from neutron.services.qos import qos_consts
@@ -56,7 +55,8 @@ OvnPortInfo = collections.namedtuple('OvnPortInfo', ['type', 'options',
                                                      'addresses',
                                                      'port_security',
                                                      'parent_name', 'tag',
-                                                     'dhcpv4_options'])
+                                                     'dhcpv4_options',
+                                                     'dhcpv6_options'])
 
 
 class OVNMechanismDriver(driver_api.MechanismDriver):
@@ -104,7 +104,7 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
     @property
     def _plugin(self):
         if self._plugin_property is None:
-            self._plugin_property = manager.NeutronManager.get_plugin()
+            self._plugin_property = directory.get_plugin()
         return self._plugin_property
 
     def _get_attribute(self, obj, attribute):
@@ -115,20 +115,17 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
 
     def _setup_vif_port_bindings(self):
         self.supported_vnic_types = [portbindings.VNIC_NORMAL]
-        # NOTE(rtheis): Config for vif_type will ensure valid choices.
-        if config.get_ovn_vif_type() == portbindings.VIF_TYPE_VHOST_USER:
-            self.vif_type = portbindings.VIF_TYPE_VHOST_USER
-            self.vif_details = {
+        self.vif_details = {
+            portbindings.VIF_TYPE_OVS: {
+                portbindings.CAP_PORT_FILTER: self.sg_enabled
+            },
+            portbindings.VIF_TYPE_VHOST_USER: {
                 portbindings.CAP_PORT_FILTER: False,
                 portbindings.VHOST_USER_MODE:
                 portbindings.VHOST_USER_MODE_CLIENT,
-                portbindings.VHOST_USER_OVS_PLUG: True,
-            }
-        else:
-            self.vif_type = portbindings.VIF_TYPE_OVS,
-            self.vif_details = {
-                portbindings.CAP_PORT_FILTER: self.sg_enabled,
+                portbindings.VHOST_USER_OVS_PLUG: True
             }
+        }
 
     def subscribe(self):
         registry.subscribe(self.post_fork_initialize,
@@ -291,14 +288,14 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                 lswitch_name=lswitch_name,
                 external_ids=ext_ids))
             if physnet:
-                vlan_id = None
+                vlan_id = []
                 if segid is not None:
                     vlan_id = int(segid)
                 txn.add(self._nb_ovn.create_lswitch_port(
                     lport_name='provnet-%s' % network['id'],
                     lswitch_name=lswitch_name,
                     addresses=['unknown'],
-                    external_ids=None,
+                    external_ids={},
                     type='localnet',
                     tag=vlan_id,
                     options={'network_name': physnet}))
@@ -391,6 +388,12 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
 
     def add_subnet_dhcp_options_in_ovn(self, subnet, network,
                                        ovn_dhcp_options=None):
+        # Don't insert DHCP_Options entry for v6 subnet with 'SLAAC' as
+        # 'ipv6_address_mode', since DHCPv6 shouldn't work for this mode.
+        if (subnet['ip_version'] == const.IP_VERSION_6 and
+                subnet.get('ipv6_address_mode') == const.IPV6_SLAAC):
+            return
+
         if not ovn_dhcp_options:
             ovn_dhcp_options = self.get_ovn_dhcp_options(subnet, network)
 
@@ -405,9 +408,12 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         dhcp_options = {'cidr': subnet['cidr'], 'options': {},
                         'external_ids': external_ids}
 
-        if subnet['ip_version'] == 4 and subnet['enable_dhcp']:
-            dhcp_options['options'] = self._get_ovn_dhcpv4_opts(
-                subnet, network, server_mac=server_mac)
+        if subnet['enable_dhcp']:
+            if subnet['ip_version'] == const.IP_VERSION_4:
+                dhcp_options['options'] = self._get_ovn_dhcpv4_opts(
+                    subnet, network, server_mac=server_mac)
+            else:
+                dhcp_options['options'] = self._get_ovn_dhcpv6_opts(subnet)
 
         return dhcp_options
 
@@ -431,11 +437,7 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                 cfg.CONF.base_mac.split(':'))
 
         if subnet['dns_nameservers']:
-            dns_servers = '{'
-            for dns in subnet["dns_nameservers"]:
-                dns_servers += dns + ', '
-            dns_servers = dns_servers.strip(', ')
-            dns_servers += '}'
+            dns_servers = '{%s}' % ', '.join(subnet['dns_nameservers'])
             options['dns_server'] = dns_servers
 
         # If subnet hostroutes are defined, add them in the
@@ -455,6 +457,22 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
 
         return options
 
+    def _get_ovn_dhcpv6_opts(self, subnet):
+        """Returns the DHCPv6 options"""
+
+        dhcpv6_opts = {
+            'server_id': n_utils.get_random_mac(cfg.CONF.base_mac.split(':'))
+        }
+
+        if subnet['dns_nameservers']:
+            dns_servers = '{%s}' % ', '.join(subnet['dns_nameservers'])
+            dhcpv6_opts['dns_server'] = dns_servers
+
+        if subnet.get('ipv6_address_mode') == const.DHCPV6_STATELESS:
+            dhcpv6_opts[ovn_const.DHCPV6_STATELESS_OPT] = 'true'
+
+        return dhcpv6_opts
+
     def create_port_precommit(self, context):
         """Allocate resources for a new port.
 
@@ -467,7 +485,8 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         """
         port = context.current
         self.validate_and_get_data_from_binding_profile(port)
-        self._insert_port_provisioning_block(context._plugin_context, port)
+        if self._is_port_provisioning_required(port, context.host):
+            self._insert_port_provisioning_block(context._plugin_context, port)
 
     def validate_and_get_data_from_binding_profile(self, port):
         if (ovn_const.OVN_PORT_BINDING_PROFILE not in port or
@@ -525,21 +544,46 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
 
         return param_dict
 
-    def _insert_port_provisioning_block(self, context, port):
+    def _is_port_provisioning_required(self, port, host, original_host=None):
         vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
         if vnic_type not in self.supported_vnic_types:
-            LOG.debug("No provisioning block due to unsupported vnic_type: %s",
-                      vnic_type)
-            return
+            LOG.debug('No provisioning block for port %(port_id)s due to '
+                      'unsupported vnic_type: %(vnic_type)s',
+                      {'port_id': port['id'], 'vnic_type': vnic_type})
+            return False
+
+        if port['status'] == const.PORT_STATUS_ACTIVE:
+            LOG.debug('No provisioning block for port %s since it is active',
+                      port['id'])
+            return False
+
+        if not host:
+            LOG.debug('No provisioning block for port %s since it does not '
+                      'have a host', port['id'])
+            return False
+
+        if host == original_host:
+            LOG.debug('No provisioning block for port %s since host unchanged',
+                      port['id'])
+            return False
+
+        if not self._sb_ovn.chassis_exists(host):
+            LOG.debug('No provisioning block for port %(port_id)s since no '
+                      'OVN chassis for host: %(host)s',
+                      {'port_id': port['id'], 'host': host})
+            return False
+
+        return True
+
+    def _insert_port_provisioning_block(self, context, port):
         # Insert a provisioning block to prevent the port from
         # transitioning to active until OVN reports back that
         # the port is up.
-        if port['status'] != const.PORT_STATUS_ACTIVE:
-            provisioning_blocks.add_provisioning_component(
-                context,
-                port['id'], resources.PORT,
-                provisioning_blocks.L2_AGENT_ENTITY
-            )
+        provisioning_blocks.add_provisioning_component(
+            context,
+            port['id'], resources.PORT,
+            provisioning_blocks.L2_AGENT_ENTITY
+        )
 
     def create_port_postcommit(self, context):
         """Create a port.
@@ -559,6 +603,9 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         if not port.get(psec.PORTSECURITY):
             return []
 
+        if utils.is_lsp_trusted(port):
+            return []
+
         allowed_addresses = set()
         addresses = port['mac_address']
         for ip in port.get('fixed_ips', []):
@@ -584,9 +631,6 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         if qos_options is None:
             qos_options = self.qos_driver.get_qos_options(port)
         vtep_physical_switch = binding_profile.get('vtep-physical-switch')
-        parent_name = None
-        tag = None
-        port_type = None
 
         if vtep_physical_switch:
             vtep_logical_switch = binding_profile.get('vtep-logical-switch')
@@ -594,23 +638,24 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
             options = {'vtep-physical-switch': vtep_physical_switch,
                        'vtep-logical-switch': vtep_logical_switch}
             addresses = "unknown"
+            parent_name = []
+            tag = []
             port_security = []
         else:
             options = qos_options
-            parent_name = binding_profile.get('parent_name')
-            tag = binding_profile.get('tag')
+            parent_name = binding_profile.get('parent_name', [])
+            tag = binding_profile.get('tag', [])
             addresses = port['mac_address']
             for ip in port.get('fixed_ips', []):
                 addresses += ' ' + ip['ip_address']
             port_security = self._get_allowed_addresses_from_port(port)
+            port_type = ''
 
-        port_dhcpv4_options_info = self.get_port_dhcpv4_options(port)
-        dhcpv4_options = []
-        if port_dhcpv4_options_info and 'uuid' in port_dhcpv4_options_info:
-            dhcpv4_options = [port_dhcpv4_options_info['uuid']]
+        dhcpv4_options = self.get_port_dhcp_options(port, const.IP_VERSION_4)
+        dhcpv6_options = self.get_port_dhcp_options(port, const.IP_VERSION_6)
 
         return OvnPortInfo(port_type, options, [addresses], port_security,
-                           parent_name, tag, dhcpv4_options)
+                           parent_name, tag, dhcpv4_options, dhcpv6_options)
 
     def create_port_in_ovn(self, port, ovn_port_info):
         external_ids = {ovn_const.OVN_PORT_NAME_EXT_ID_KEY: port['name']}
@@ -619,7 +664,28 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         sg_cache = {}
         subnet_cache = {}
 
+        # It's possible to have a network created on one controller and then a
+        # port created on a different controller quickly enough that the second
+        # controller does not yet see that network in its local cache of the
+        # OVN northbound database.  Check if the logical switch is present
+        # or not in the idl's local copy of the database before creating
+        # the lswitch port.
+        self._nb_ovn.check_for_row_by_value_and_retry(
+            'Logical_Switch', 'name', lswitch_name)
+
         with self._nb_ovn.transaction(check_error=True) as txn:
+            if not ovn_port_info.dhcpv4_options:
+                dhcpv4_options = []
+            elif 'cmd' in ovn_port_info.dhcpv4_options:
+                dhcpv4_options = txn.add(ovn_port_info.dhcpv4_options['cmd'])
+            else:
+                dhcpv4_options = [ovn_port_info.dhcpv4_options['uuid']]
+            if not ovn_port_info.dhcpv6_options:
+                dhcpv6_options = []
+            elif 'cmd' in ovn_port_info.dhcpv6_options:
+                dhcpv6_options = txn.add(ovn_port_info.dhcpv6_options['cmd'])
+            else:
+                dhcpv6_options = [ovn_port_info.dhcpv6_options['uuid']]
             # The lport_name *must* be neutron port['id'].  It must match the
             # iface-id set in the Interfaces table of the Open_vSwitch
             # database which nova sets to be the port ID.
@@ -634,14 +700,15 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                     options=ovn_port_info.options,
                     type=ovn_port_info.type,
                     port_security=ovn_port_info.port_security,
-                    dhcpv4_options=ovn_port_info.dhcpv4_options))
+                    dhcpv4_options=dhcpv4_options,
+                    dhcpv6_options=dhcpv6_options))
 
             acls_new = ovn_acl.add_acls(self._plugin, admin_context,
                                         port, sg_cache, subnet_cache)
             for acl in acls_new:
                 txn.add(self._nb_ovn.add_acl(**acl))
 
-            sg_ids = port.get('security_groups', [])
+            sg_ids = utils.get_lsp_security_groups(port)
             if port.get('fixed_ips') and sg_ids:
                 addresses = ovn_acl.acl_port_ips(port)
                 # NOTE(rtheis): Fail port creation if the address set doesn't
@@ -671,7 +738,11 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         state. It is up to the mechanism driver to ignore state or
         state changes that it does not know or care about.
         """
-        self.validate_and_get_data_from_binding_profile(context.current)
+        port = context.current
+        self.validate_and_get_data_from_binding_profile(port)
+        if self._is_port_provisioning_required(port, context.host,
+                                               context.original_host):
+            self._insert_port_provisioning_block(context._plugin_context, port)
 
     def update_port_postcommit(self, context):
         """Update a port.
@@ -705,21 +776,47 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         subnet_cache = {}
 
         with self._nb_ovn.transaction(check_error=True) as txn:
+            columns_dict = {}
+            if port.get('device_owner') in [const.DEVICE_OWNER_ROUTER_INTF,
+                                            const.DEVICE_OWNER_ROUTER_GW]:
+                ovn_port_info.options.update(
+                    self._nb_ovn.get_router_port_options(port['id']))
+            else:
+                columns_dict['type'] = ovn_port_info.type
+            if not ovn_port_info.dhcpv4_options:
+                dhcpv4_options = []
+            elif 'cmd' in ovn_port_info.dhcpv4_options:
+                dhcpv4_options = txn.add(ovn_port_info.dhcpv4_options['cmd'])
+            else:
+                dhcpv4_options = [ovn_port_info.dhcpv4_options['uuid']]
+            if not ovn_port_info.dhcpv6_options:
+                dhcpv6_options = []
+            elif 'cmd' in ovn_port_info.dhcpv6_options:
+                dhcpv6_options = txn.add(ovn_port_info.dhcpv6_options['cmd'])
+            else:
+                dhcpv6_options = [ovn_port_info.dhcpv6_options['uuid']]
+            # NOTE(lizk): Fail port updating if port doesn't exist. This
+            # prevents any new inserted resources to be orphan, such as port
+            # dhcp options or ACL rules for port, e.g. a port was created
+            # without extra dhcp options and security group, while updating
+            # includes the new attributes setting to port.
             txn.add(self._nb_ovn.set_lswitch_port(
                     lport_name=port['id'],
                     addresses=ovn_port_info.addresses,
                     external_ids=external_ids,
                     parent_name=ovn_port_info.parent_name,
                     tag=ovn_port_info.tag,
-                    type=ovn_port_info.type,
                     options=ovn_port_info.options,
                     enabled=port['admin_state_up'],
                     port_security=ovn_port_info.port_security,
-                    dhcpv4_options=ovn_port_info.dhcpv4_options))
+                    dhcpv4_options=dhcpv4_options,
+                    dhcpv6_options=dhcpv6_options,
+                    if_exists=False,
+                    **columns_dict))
 
             # Determine if security groups or fixed IPs are updated.
-            old_sg_ids = set(original_port.get('security_groups', []))
-            new_sg_ids = set(port.get('security_groups', []))
+            old_sg_ids = set(utils.get_lsp_security_groups(original_port))
+            new_sg_ids = set(utils.get_lsp_security_groups(port))
             detached_sg_ids = old_sg_ids - new_sg_ids
             attached_sg_ids = new_sg_ids - old_sg_ids
             is_fixed_ips_updated = \
@@ -781,73 +878,75 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                                         addrs_add=addr_add,
                                         addrs_remove=addr_remove))
 
-    def _get_delete_lsp_dhcpv4_options_cmd(self, port):
-        lsp_dhcp_options = None
-        for fixed_ip in port['fixed_ips']:
-            if netaddr.IPAddress(fixed_ip['ip_address']).version == 4:
-                lsp_dhcp_options = self._nb_ovn.get_port_dhcp_options(
-                    fixed_ip['subnet_id'], port['id'])
-                if lsp_dhcp_options:
-                    break
+    def _get_subnet_dhcp_options_for_port(self, port, ip_version):
+        """Returns the subnet dhcp options for the port.
 
-        if lsp_dhcp_options:
-            # Extra DHCP options were defined for this port. Delete the
-            # DHCP_Options row created for this port earlier if exists,
-            # since this port no longer refers it.
-            return self._nb_ovn.delete_dhcp_options(lsp_dhcp_options['uuid'])
+        Return the first found DHCP options belong for the port.
+        """
+        subnets = [
+            fixed_ip['subnet_id']
+            for fixed_ip in port['fixed_ips']
+            if netaddr.IPAddress(fixed_ip['ip_address']).version == ip_version]
+        get_opts = self._nb_ovn.get_subnets_dhcp_options(subnets)
+        if get_opts:
+            if ip_version == const.IP_VERSION_6:
+                # Always try to find a dhcpv6 stateful v6 subnet to return.
+                # This ensures port can get one stateful v6 address when port
+                # has multiple dhcpv6 stateful and stateless subnets.
+                for opts in get_opts:
+                    # We are setting ovn_const.DHCPV6_STATELESS_OPT to "true"
+                    # in _get_ovn_dhcpv6_opts, so entries in DHCP_Options table
+                    # should have unicode type 'true' if they were defined as
+                    # dhcpv6 stateless.
+                    if opts['options'].get(
+                        ovn_const.DHCPV6_STATELESS_OPT) != 'true':
+                        return opts
+            return get_opts[0]
 
-    def get_port_dhcpv4_options(self, port):
-        lsp_dhcp_disabled, lsp_dhcpv4_opts = utils.get_lsp_dhcpv4_opts(port)
+    def get_port_dhcp_options(self, port, ip_version):
+        """Return dhcp options for port.
+
+        In case the port is dhcp disabled, or IP addresses it has belong
+        to dhcp disabled subnets, returns None.
+        Otherwise, returns a dict:
+         - with content from a existing DHCP_Options row for subnet, if the
+           port has no extra dhcp optoins.
+         - with only one item ('cmd', AddDHCPOptionsCommand(..)), if the port
+           has extra dhcp options. The command should be processed in the same
+           transaction with port creating or updating command to avoid orphan
+           row issue happen.
+        """
+        lsp_dhcp_disabled, lsp_dhcp_opts = utils.get_lsp_dhcp_opts(
+            port, ip_version)
 
         if lsp_dhcp_disabled:
             return
 
-        # If the port has multiple IPv4 addresses, DHCPv4 options are set
-        # for the first address in port['fixed_ips']
-        subnet_dhcp_options = None
-        subnet_id = None
-        for fixed_ip in port['fixed_ips']:
-            if netaddr.IPAddress(fixed_ip['ip_address']).version == 4:
-                subnet_dhcp_options = self._nb_ovn.get_subnet_dhcp_options(
-                    fixed_ip['subnet_id'])
-                subnet_id = fixed_ip['subnet_id']
-                if subnet_dhcp_options:
-                    break
+        subnet_dhcp_options = self._get_subnet_dhcp_options_for_port(
+            port, ip_version)
 
         if not subnet_dhcp_options:
-            # Ideally this should not happen.
-            # May be a sync is required in such cases ?
+            # NOTE(lizk): It's possible for Neutron to configure a port with IP
+            # address belongs to subnet disabled dhcp. And no DHCP_Options row
+            # wll be inserted for such a subnet. So in that case, the subnet
+            # dhcp options here will be None.
             return
 
-        if not lsp_dhcpv4_opts:
+        if not lsp_dhcp_opts:
             return subnet_dhcp_options
 
-        # This port has extra DHCP options defined.
-        # So we need to create a new row in DHCP_Options table for this
-        # port.
-        #
-        # TODO(numans) In cases where the below transaction is successful
-        # but the Logical_Switch_Port create or update transaction fails
-        # we need to delete the DHCP_Options row created else it will be
-        # an orphan row.
-        #
-        # NOTE(lizk) In cases where the below transaction is successful, but
-        # the Logical_Switch_Port get deleted before setting port dhcp options
-        # to it, we will delete the DHCP_Options row created to make sure
-        # no orphan left behind.
-        subnet_dhcp_options['options'].update(lsp_dhcpv4_opts)
+        # This port has extra DHCP options defined, so we will create a new
+        # row in DHCP_Options table for it.
+        subnet_dhcp_options['options'].update(lsp_dhcp_opts)
         subnet_dhcp_options['external_ids'].update(
             {'port_id': port['id']})
-        LOG.debug('Creating port dhcp options for port %s in OVN NB DB',
-                  port['id'])
-        with self._nb_ovn.transaction(check_error=True) as txn:
-            txn.add(self._nb_ovn.add_dhcp_options(
-                subnet_id, port_id=port['id'],
-                cidr=subnet_dhcp_options['cidr'],
-                options=subnet_dhcp_options['options'],
-                external_ids=subnet_dhcp_options['external_ids']))
-
-        return self._nb_ovn.get_port_dhcp_options(subnet_id, port['id'])
+        subnet_id = subnet_dhcp_options['external_ids']['subnet_id']
+        add_dhcp_opts_cmd = self._nb_ovn.add_dhcp_options(
+            subnet_id, port_id=port['id'],
+            cidr=subnet_dhcp_options['cidr'],
+            options=subnet_dhcp_options['options'],
+            external_ids=subnet_dhcp_options['external_ids'])
+        return {'cmd': add_dhcp_opts_cmd}
 
     def delete_port_postcommit(self, context):
         """Delete a port.
@@ -870,7 +969,8 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
 
             if port.get('fixed_ips'):
                 addresses = ovn_acl.acl_port_ips(port)
-                for sg_id in port.get('security_groups', []):
+                # Set skip_trusted_port False for deleting port
+                for sg_id in utils.get_lsp_security_groups(port, False):
                     for ip_version in addresses:
                         if addresses[ip_version]:
                             txn.add(self._nb_ovn.update_address_set(
@@ -878,13 +978,6 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                                 addrs_add=None,
                                 addrs_remove=addresses[ip_version]))
 
-            # NOTE(lizk): Always try to clean port dhcp options, to make sure
-            # no orphaned DHCP_Options row related to port left behind, which
-            # may be created in get_port_dhcpv4_options.
-            cmd = self._get_delete_lsp_dhcpv4_options_cmd(port)
-            if cmd:
-                txn.add(cmd)
-
     def bind_port(self, context):
         """Attempt to bind a port.
 
@@ -934,15 +1027,30 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                       {'port_id': port['id'], 'vnic_type': vnic_type})
             return
 
+        # OVN chassis information is needed to ensure a valid port bind.
+        # Collect port binding data and refuse binding if the OVN chassis
+        # cannot be found.
+        chassis_physnets = []
+        try:
+            datapath_type, iface_types, chassis_physnets = \
+                self._sb_ovn.get_chassis_data_for_ml2_bind_port(context.host)
+            iface_types = iface_types.split(',') if iface_types else []
+        except RuntimeError:
+            LOG.debug('Refusing to bind port %(port_id)s due to '
+                      'no OVN chassis for host: %(host)s' %
+                      {'port_id': port['id'], 'host': context.host})
+            return
+
         for segment_to_bind in context.segments_to_bind:
             network_type = segment_to_bind['network_type']
             segmentation_id = segment_to_bind['segmentation_id']
             physical_network = segment_to_bind['physical_network']
-            LOG.debug('Attempting to bind port %(port_id)s for network '
-                      'segment with type %(network_type)s, '
+            LOG.debug('Attempting to bind port %(port_id)s on host %(host)s '
+                      'for network segment with type %(network_type)s, '
                       'segmentation ID %(segmentation_id)s, '
                       'physical network %(physical_network)s' %
                       {'port_id': port['id'],
+                       'host': context.host,
                        'network_type': network_type,
                        'segmentation_id': segmentation_id,
                        'physical_network': physical_network})
@@ -956,14 +1064,36 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
                              'unsupported network type: %(network_type)s'),
                          {'port_id': port['id'],
                           'network_type': network_type})
-            if self.vif_type == portbindings.VIF_TYPE_VHOST_USER:
-                port[portbindings.VIF_DETAILS].update({
-                    portbindings.VHOST_USER_SOCKET: utils.ovn_vhu_sockpath(
-                        cfg.CONF.ovn.vhost_sock_dir, port['id'])
-                    })
-            context.set_binding(segment_to_bind[driver_api.ID],
-                                self.vif_type,
-                                self.vif_details)
+
+            if (network_type in ['flat', 'vlan']) and \
+               (physical_network not in chassis_physnets):
+                LOG.info(_LI('Refusing to bind port %(port_id)s on '
+                             'host %(host)s due to the OVN chassis '
+                             'bridge mapping physical networks '
+                             '%(chassis_physnets)s not supporting '
+                             'physical network: %(physical_network)s'),
+                         {'port_id': port['id'],
+                          'host': context.host,
+                          'chassis_physnets': chassis_physnets,
+                          'physical_network': physical_network})
+            else:
+                if datapath_type == ovn_const.CHASSIS_DATAPATH_NETDEV and (
+                    ovn_const.CHASSIS_IFACE_DPDKVHOSTUSER in iface_types):
+                    vhost_user_socket = utils.ovn_vhu_sockpath(
+                        config.get_ovn_vhost_sock_dir(), port['id'])
+                    vif_type = portbindings.VIF_TYPE_VHOST_USER
+                    port[portbindings.VIF_DETAILS].update({
+                        portbindings.VHOST_USER_SOCKET: vhost_user_socket
+                        })
+                    vif_details = dict(self.vif_details[vif_type])
+                    vif_details[portbindings.VHOST_USER_SOCKET] = \
+                        vhost_user_socket
+                else:
+                    vif_type = portbindings.VIF_TYPE_OVS
+                    vif_details = self.vif_details[vif_type]
+
+                context.set_binding(segment_to_bind[driver_api.ID], vif_type,
+                                    vif_details)
 
     def get_workers(self):
         """Get any NeutronWorker instances that should have their own process
@@ -995,7 +1125,6 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
         admin_context = n_context.get_admin_context()
         try:
             port = self._plugin.get_port(admin_context, port_id)
-            port['status'] = const.PORT_STATUS_DOWN
             self._insert_port_provisioning_block(admin_context, port)
             self._plugin.update_port_status(admin_context,
                                             port['id'],
@@ -1027,6 +1156,6 @@ class OVNMechanismDriver(driver_api.MechanismDriver):
             return
 
         host_phynets_map = self._sb_ovn.get_chassis_hostname_and_physnets()
-        hosts = {host for host, phynets in six.iteritems(host_phynets_map)
+        hosts = {host for host, phynets in host_phynets_map.items()
                  if phynet in phynets}
         segment_service_db.map_segment_to_hosts(context, segment.id, hosts)
